[trial, settingsEval] = Experiments.getTrialForScript();
close all;

if (trial.isConfigure)
    
    
    % Experiments settings
    settings = Common.Settings();
    settings.setProperty('numParameters', 15);
    settings.setProperty('numContexts', 0);
    settings.setProperty('numSamplesEpisodes', 15);
    settings.setProperty('numIterations', 200);
    settings.setProperty('useImportanceWeighting', false);
    settings.setProperty('maxSamples', 200);
    settings.setProperty('numInitialSamplesEpisodes',100);
    
    settings.setProperty('epsilonAction',0.5);
    settings.setProperty('EpisodicREPSOptimizationOptiAlgorithm', 'FMinCon');
    
    
    
    trial.configure(settingsEval);
    
    
    
    % create steps sampler
    sampler = Sampler.EpisodeSampler();
    dataManager = sampler.getDataManager();
    
    %define environment
    returnSampler = Environments.BanditEnvironments.CigarReward(sampler, settings.numContexts, settings.numParameters);
    
    parameterPolicy = Distributions.Gaussian.GaussianParameterPolicy(dataManager);
    
    importanceWeighting = DataPreprocessors.ImportanceSamplingLastKPolicies(dataManager, parameterPolicy);
        
    parameterPolicyLearner = Learner.SupervisedLearner.LinearGaussianMLLearner(dataManager, parameterPolicy);
    
    policyLearner = Learner.EpisodicRL.EpisodicREPS(dataManager, parameterPolicyLearner, 'returns', 'returnWeightings');   
    
    policyLearner.initObject();
    
    if (settings.useImportanceWeighting)
        policyLearner.setWeightName(importanceWeighting.importanceWeightName);
    end
    
    
    sampler.setParameterPolicy(parameterPolicy);
    
    sampler.setReturnFunction(returnSampler);
    
    deletionStrategy = LearningScenario.MaxSamplesDeletionStrategy();
    
    dataManager.finalizeDataManager();
end

if (trial.isStart)
    newData = dataManager.getDataObject(0);
    data = dataManager.getDataObject(0);
    
    parameterPolicy.initObject();
    
    for i = 1:settings.numIterations
        
        sampler.setSamplerIteration(i);
        sampler.createSamples(newData);
        
        % keep old samples strategy comes here...
        data.mergeData(newData);
        deletionStrategy.deleteSamples(data);
        
        % data preprocessors come here
        % ...
        importanceWeighting.preprocessData(data);
        
        % learning comes here...
        policyLearner.updateModel(data);
        
        % log the results...
        trial.store('avgReturns', mean(newData.getDataEntry('returns')), Experiments.StoringType.ACCUMULATE)
        %trial.store('entropy', policyLearner.entropyAfter, Experiments.StoringType.ACCUMULATE);
        %trial.store('divMean', policyLearner.divMean, Experiments.StoringType.ACCUMULATE);
        %trial.store('divCov', policyLearner.divCov, Experiments.StoringType.ACCUMULATE);
        trial.store('divKL', policyLearner.divKL, Experiments.StoringType.ACCUMULATE);
        
        
        
        trial.saveWorkspace();
        trial.storeTrial();
        
        % fprintf('Iteration %d: %f, divMean: %f divCov: %f \n',i, trial.avgReturns(end, 1), trial.divMean(end) , trial.divCov(end));
        fprintf('Iteration %d: %f, divKL: %f \n',i, trial.avgReturns(end, 1), trial.divKL(end));
        %fprintf('Iteration %d: %f \n',i, trial.avgReturns(end, 1));
        
    end
end