[trial, settingsEval] = Experiments.getTrialForScript();
close all;

if (trial.isConfigure)
    
    % Experiments settings
    settings = Common.Settings();
    settings.setProperty('numParameters', 15);
    settings.setProperty('numContexts', 0);
    settings.setProperty('numSamplesEpisodes', 15);
    settings.setProperty('numIterations', 3000);
    settings.setProperty('L', 15);
    
    
    trial.configure(settingsEval);
    
    % create steps sampler
    sampler = Sampler.EpisodeSampler();
    dataManager = sampler.getDataManager();
    
    %define environment
    returnSampler = Environments.BanditEnvironments.RastriginReward(sampler, settings.numContexts, settings.numParameters);
    
    parameterPolicy = Distributions.Gaussian.GaussianParameterPolicy(dataManager);
    policyLearner = Learner.EpisodicRL.NESLearner2(dataManager, parameterPolicy);
    
    sampler.setParameterPolicy(parameterPolicy);
    
    sampler.setReturnFunction(returnSampler);
    
    dataManager.finalizeDataManager();
end

if (trial.isStart)
    newData = dataManager.getDataObject(0);
    
    parameterPolicy.initObject();
    
    for i = 1:settings.numIterations
        
        sampler.setSamplerIteration(i);
        
        sampler.createSamples(newData);
        
        % keep old samples strategy comes here...
        data = newData;
        
        % data preprocessors come here
        % ...
        
        policyLearner.updateModel(newData);
        
        trial.store('avgReturns', mean(newData.getDataEntry('returns')), Experiments.StoringType.ACCUMULATE)
        
        fprintf('Iteration: %d, Episodes: %d, Avg Return: %f\n', i, i * settings.numSamplesEpisodes, mean(newData.getDataEntry('returns')));
    end
    
end