[trial, settingsEval] = Experiments.getTrialForScript(); 
close all;

if (trial.isConfigure)
    

    % Experiments settings
    settings = Common.Settings();
    settings.setProperty('numParameters', 15);
    settings.setProperty('numContexts', 0);
    settings.setProperty('numSamplesEpisodes', 15);   
    settings.setProperty('numIterations', 850);
    settings.setProperty('useImportanceWeighting', false);
    settings.setProperty('maxSamples', 200);
    settings.setProperty('numInitialSamplesEpisodes',100);
    
    settings.setProperty('entropyBetaDiscount',0.18);
    settings.setProperty('epsilonAction',3);
    settings.setProperty('minEntropy',-5000);
    settings.setProperty('entropyUpdateType','absolute');



    trial.configure(settingsEval);

        

    % create steps sampler
    sampler = Sampler.EpisodeSampler();
    dataManager = sampler.getDataManager();
        
    %define environment
    returnSampler = Environments.BanditEnvironments.RosenbrockReward(sampler, settings.numContexts, settings.numParameters);

    parameterPolicy = Distributions.Gaussian.GaussianParameterPolicy(dataManager);
    
    quadraticModel = Functions.SquaredFunction(dataManager, 'returns', 'parameters', 'rewardModel');

    quadraticModelLearner = Learner.ModelLearner.ConvexQuadraticModelLearner(dataManager, quadraticModel);

    importanceWeighting = DataPreprocessors.ImportanceSamplingLastKPolicies(dataManager, parameterPolicy);
    
    if (settings.useImportanceWeighting)
        quadraticModelLearner.setWeightName(importanceWeighting.importanceWeightName);
    end    
    
    policyLearner = Learner.EpisodicRL.ConstraintEntropyREPSCLosedForm(dataManager, parameterPolicy, quadraticModel, quadraticModelLearner, [], []);
    policyLearner.initObject();
    
    sampler.setParameterPolicy(parameterPolicy);

    sampler.setReturnFunction(returnSampler);
        
    deletionStrategy = LearningScenario.MaxSamplesDeletionStrategy();

    dataManager.finalizeDataManager();
end

if (trial.isStart)
    newData = dataManager.getDataObject(0);
    data = dataManager.getDataObject(0);

    parameterPolicy.initObject();
    
    for i = 1:settings.numIterations
        
        sampler.setSamplerIteration(i);

        sampler.createSamples(newData);
        
        % keep old samples strategy comes here...
        data.mergeData(newData);
        deletionStrategy.deleteSamples(data);
        
        % data preprocessors come here
        % ...
        importanceWeighting.preprocessData(data);
        
        % learning comes here...
        policyLearner.updateModel(data);
        
        % log the results...
        trial.store('avgReturns', mean(newData.getDataEntry('returns')), Experiments.StoringType.ACCUMULATE)
        %trial.store('entropy', policyLearner.entropyAfter, Experiments.StoringType.ACCUMULATE);
        trial.store('divMean', policyLearner.divMean, Experiments.StoringType.ACCUMULATE);
        trial.store('divCov', policyLearner.divCov, Experiments.StoringType.ACCUMULATE);
        %trial.store('divKL', policyLearner.divKL, Experiments.StoringType.ACCUMULATE);

        
        
       % trial.saveWorkspace();
       % trial.storeTrial();
        
        fprintf('Iteration %d, Average Reward: %f, NumSamples: %f\n', i, mean(newData.getDataEntry('returns')), data.getNumElements());
        policyLearner.printMessage(data)
        %fprintf('Iteration %d: %f, divKL: %f \n',i, trial.avgReturns(end, 1), trial.divKL(end));
        %fprintf('Iteration %d: %f \n',i, trial.avgReturns(end, 1));

    end
end