classdef ConstraintEntropyREPSCLosedForm < Learner.RLLearner & Data.DataManipulator & Data.OptionalParameterInterface
    
    properties (SetObservable, AbortSet)
        entropyBeta = 1.0;
        entropyBetaDiscount = 1;
        numOptimizationsDualFunction = 1000;
        epsilonAction = 1.5;
        minEntropy = -1000;
        
        entropyUpdateType = 'absolute';
        
        
    end
    
    properties
        useGradient;
        divKL
        divMean
        divCov
        
        stateDistribution = [];
        rewardModel = [];
        
        stateDistributionLearner = [];
        rewardModelLearner = [];
        policy = [];
        entropyBefore;
        entropyAfter;
        
        Ras;
        Raa;
        Rss;
        ra;
        rs;
        r0;
        
        bias;
        K;
        Q;
        Qinv;
        logDetQ;
        q;
        
        eta = 1;
        gammA = 1;
        
        muState;
        SigmaState;
        
        posDefVal = false;
        
        rewardBeforeUpdate
        rewardAfterUpdate
        
        desiredEntropy
        
        optimizer;
        
    end
    
    methods (Static)
        function [learner] = CreateFromTrial(trial)
            learner = Learner.EpisodicRL.ConstraintEntropyREPSCLosedForm(trial.dataManager, trial.parameterPolicy, ...
                trial.learnedRewardFunction, trial.rewardFunctionLearner, trial.learnedContextDistribution, trial.contextDistributionLearner);
        end
    end
    
    % Class methods
    methods
        function obj = ConstraintEntropyREPSCLosedForm(dataManager, policy, rewardModel, rewardModelLearner, stateDistribution, stateDistributionLearner)
            obj = obj@Learner.RLLearner();
            obj = obj@Data.DataManipulator(dataManager);
            
            obj.useGradient   = 'on';
            
            obj.rewardModel   = rewardModel;
            obj.rewardModelLearner = rewardModelLearner;
            
            obj.policy   = policy;
            obj.stateDistribution = stateDistribution;
            obj.stateDistributionLearner = stateDistributionLearner;
            
            %obj.useHessian    = true;
            
            obj.linkProperty('entropyBeta');
            obj.linkProperty('entropyBetaDiscount');
            obj.linkProperty('numOptimizationsDualFunction');
            obj.linkProperty('epsilonAction');
            obj.linkProperty('minEntropy');
            obj.linkProperty('entropyUpdateType');
            
            obj.settings.setIfEmpty('EpisodicREPSOptimizationNumIterations', 500);
            obj.settings.setIfEmpty('EpisodicREPSOptimizationOptiRelfTol', 10^-8);
            
        end
        
        function [] = registerHierarchicalParameters(obj, hierarchicalDataManager, parameterName)
            obj.registerOptionalParameter('minEntropy', true, 1, -20, 0, parameterName, 1, 'logsig', hierarchicalDataManager);
            obj.registerOptionalParameter('epsilonAction', false, 1, 0, 2.0, parameterName, 1, 'logsig', hierarchicalDataManager);
        end
        
        %%
        function [] = updateEntropyBeta(obj)
            
            %obj.entropyBeta = obj.eta * obj.entropyBetaDiscount ;
            %cov = obj.policy.getCovariance();
            bias = obj.policy.bias;
            %obj.entropyBeta =  0.5*length(bias)*(1+log(2*pi)) + 0.5*sum(log(eig(cov))); %1/2*(log(((2*pi*exp(1))^length(bias)) * det(cov))/log(exp(1)));
            
            cholCov = obj.policy.cholA();
            v = 2*sum(log(diag(cholCov))); %lodDet Policy covariance
            
            obj.entropyBeta =0.5*length(bias)*(1+log(2*pi)) + 0.5*v; %1/2*(log(((2*pi*exp(1))^length(bias)) * det(cov))/log(exp(1)));
            
            
        end
        
        
        %%
        function obj = initObject(obj)
            
            obj.eta =1000;
            obj.gammA = 1;
            params = [obj.gammA , obj.eta];
            
            [lowerBound, upperBound] = obj.getBoundsForOptimization();
            
            obj.optimizer = Optimizer.OptimizerFactory.createOptimizer(numel(params), 'EpisodicREPSOptimization', ...
                lowerBound, upperBound,'FMinCon');
            obj.initObject@Learner.RLLearner();
        end
        
        function [] = computeDesiredEntropy(obj)
            bias = obj.policy.bias;
            %obj.entropyBeta =  0.5*length(bias)*(1+log(2*pi)) + 0.5*sum(log(eig(cov))); %1/2*(log(((2*pi*exp(1))^length(bias)) * det(cov))/log(exp(1)));
            
            cholCov = obj.policy.cholA();
            v = 2*sum(log(diag(cholCov))); %lodDet Policy covariance
            %500
            currentEntropy = 0.5*length(bias)*(1+log(2*pi)) + 0.5*v; %1/2*(log(((2*pi*exp(1))^length(bias)) * det(cov))/log(exp(1)));
            obj.entropyBefore = currentEntropy;
            
            switch (obj.entropyUpdateType)
                case 'relative'
                    obj.desiredEntropy =  (obj.entropyBetaDiscount*(currentEntropy - obj.minEntropy) + obj.minEntropy) ;
                    %if(currentEntropy - obj.minEntropy <= 50)
                    % obj.minEntropy = obj.minEntropy -0.5;
                    %end
                    
                case 'absolute'
                    entropyDelta = obj.minEntropy - currentEntropy;
                    if (entropyDelta > obj.entropyBetaDiscount)
                        entropyDelta = obj.entropyBetaDiscount;
                    end
                    if (entropyDelta < - obj.entropyBetaDiscount)
                        entropyDelta = -obj.entropyBetaDiscount;
                    end
                    obj.desiredEntropy = currentEntropy + entropyDelta;
            end
            
            
        end
        
        function [] = updateModel(obj, data)
            %data.rewardWeighting = data.sampleWeighting;
            %obj.model.updateModel(data);
            
            %data = obj.model.setDataProbabilities(data);
            
            obj.computeDesiredEntropy();
            
            obj.rewardModelLearner.updateModel(data);
            
            if (~isempty(obj.stateDistributionLearner))
                obj.stateDistributionLearner.updateModel(data);
                obj.muState = obj.stateDistribution.getMean();
                obj.SigmaState = obj.stateDistribution.getCovariance();
                
                [muJoint, covJoint] = obj.policy.getJointGaussians(obj.muState, obj.SigmaState);
                dimContext = obj.dataManager.getNumDimensions('contexts');
            else
                dimContext = 0;
                covJoint = obj.policy.getCovariance();
                muJoint = obj.policy.bias;
            end
            
            obj.Q = obj.policy.getCovariance();
            obj.bias = obj.policy.bias;
            obj.K = obj.policy.weights;
            
            [R, r, obj.r0] = obj.rewardModel.getParameters();
            obj.rewardBeforeUpdate = muJoint' * R * muJoint + trace(R * covJoint) + muJoint' * r' + obj.r0;
            
            
            dimParameters = obj.policy.dimOutput;
            
            %[U, D] = eig(R);
            %D(D > 0) = 0;
            %R = U * D * U';
            
            obj.Rss = R(1:dimContext, 1:dimContext);
            obj.Ras = R(dimContext+1:end, 1:dimContext);
            obj.Raa = R(dimContext+1:end, dimContext+1:end);
            
            obj.rs = r(1:dimContext)';
            obj.ra = r(dimContext + 1:end)';
            
            obj.Qinv = inv(obj.Q);
            %obj.logDetQ = sum(log(eig(obj.Q)));
            obj.logDetQ = 2*sum(log(diag(obj.policy.cholA)));
            
            updatedPolicy = false;
            numTrys = 1;
            while (~updatedPolicy && numTrys < 5)
                obj.optimizeDualFunction(data);
                updatedPolicy = obj.updatePolicy();
                if (~updatedPolicy)
                    obj.desiredEntropy = obj.entropyBefore * 0.5 + obj.desiredEntropy * 0.5;
                end
                numTrys = numTrys + 1;
            end
            
            
        end
        
        function [] = printMessage(obj, data)
            %fprintf('Eta: %f,gammA: %f,entropyBefore: %f,entropyAfter: %f,entropyRatio: %f,divKL %f, rewardBefore: %f, rewardAfter: %f\n', obj.eta,obj.gammA,obj.entropyBeta,obj.entropyAfter,obj.entropyAfter/obj.entropyBeta,obj.divKL, obj.rewardBeforeUpdate, obj.rewardAfterUpdate);
                    fprintf('Eta: %f,gammA: %f,entropy: %f, divKL %f\n', obj.eta,obj.gammA,obj.entropyAfter,obj.divKL);

        end
        
        function [] = addDefaultCriteria(obj, trial, evaluationCriterion)
            evaluationCriterion.addCriterion('endLoop', '', 'divKL', Experiments.StoringType.ACCUMULATE, @()obj.divKL);
            evaluationCriterion.addCriterion('endLoop', '', 'rewardBefore', Experiments.StoringType.ACCUMULATE, @()obj.rewardBeforeUpdate);
            evaluationCriterion.addCriterion('endLoop', '', 'rewardAfter', Experiments.StoringType.ACCUMULATE, @()obj.rewardAfterUpdate);
        end
        
        
        function [] = optimizeDualFunction(obj, data)
            
            [params, lowerBound, upperBound] = obj.getParametersForOptimization(data);
            
            %             if(all(eig(obj.Raa)<0))
            nonlcon = [];
            %             else
            %               nonlcon = obj.getConstraintsForOptimization(data);
            %             end
            %nonlcon = [];
            %if (true)
            %               [g, gD, gDNumeric] = obj.numericDualFunctionGradient(obj.getParametersForOptimization(data), data);
            %               fprintf('Gradient Check: MAE of numeric Gradient %f\n', sum(abs(gD - gDNumeric)));
            %            end
            
            options = obj.getOptimizationOptions();
            
            dualFunctionAnonymous = @(params) dualFunction(obj, params);
            
            if (isinf(dualFunctionAnonymous(params)))
                params = [1 1];
            end
            %if(~obj.useHessian)
            %while (dualFunctionAnonymous(params) > 10^10)
            %    params = params * 2;
            %end
            for i = 1:20
                try
                    [newParams,fval,exitflag,output] = fmincon(dualFunctionAnonymous,params,[],[],[],[],lowerBound,upperBound,nonlcon, options);
                    %  [newParams,fval,exitflag]  = obj.optimizer.optimize(dualFunctionAnonymous, params);
                    if (~isinf(fval))
                        break;
                    else
                        params(2) = params(2) * 10;
                    end
                catch
                    
                    params(2) = params(2) * 10;
                    % params(1) = params(2) * 10;
                    
                end
            end
            %%%RandomSearchOpt%%%%%
            %eta = linspace(lowerBound, upperBound, 10000);
            %g = zeros(size(eta));
            %for i = 1:length(eta)
            %    g(i) = dualFunctionAnonymous(eta(i));
            %end
            %[~, ind] = min(g);
            %newParams = eta(ind);
            
            obj.setParametersFromOptimization(newParams);
            
            %else
            %              [newParams, fvalTheta] = obj.optimizeIterative(params, data, lowerBound, upperBound, options);
            %           end
            
            
        end
        
        
        
        function [lowerBound, upperBound] = getBoundsForOptimization(obj)
            
            lowerBound = [-1e20; 1e-20]';
            upperBound = [1e20; 1e20]';
        end
        %%
        function [params, lowerBound, upperBound] = getParametersForOptimization(obj, data)
            
            % obj.eta =1000;
            % obj.gammA = 1;
            params = [obj.gammA , obj.eta];
            
            %             if(obj.posDefVal)
            %                 params = [obj.eta ];
            %             else
            %                 params = [obj.eta ];
            %             end
            %
            
            %stdRew = std(data.getDataEntry('returns'));
            %if (params(end) < stdRew )
            %   params(end) = stdRew;
            %end
            
            %                if (params(1) < stdRew )
            %                  params(1) = stdRew;
            %             end
            %             lowerBound = [- ones(numel(params)-1, 1) * 1e8;  stdRew / 50];
            %             upperBound = [ones(numel(params)-1, 1) * 1e8; stdRew*50];
            %
            
            [lowerBound, upperBound] = obj.getBoundsForOptimization();
            %lowerBound = [-1e20; 1e-20]';
            %upperBound = [1e20; 1e20]';
            
        end
        
        function [nonlcon] = getConstraintsForOptimization(obj, data)
            nonlcon = @(x)obj.ensurePositiveDefiniteOmega(x);
            if(obj.posDefVal)
                nonlcon = [];
            end
        end
        
        function [c, ceq] = ensurePositiveDefiniteOmega(obj,params)
            ceq = [];
            
            %             nv = size(obj.valueSquared,1);
            %             V = zeros(nv);
            %             V(tril(true(nv))) = params((nv+1):(end-1));
            %             V = V+tril(V,-1)';
            
            eta = params(end);
            
            %            W = zeros(size(obj.Q));
            %            W(1:nv,1:nv) = V;
            
            %Omega = eta*obj.Qinv+obj.R+2*W;
            Omega = eta*obj.Qinv + 2 * obj.Raa;
            c = -eig(Omega) + 1e-12;
            %c = -sum(log(eig(Omega)))+log(1e-12)* size(obj.R,1);
            %           if(any(eig(Omega)<=0))
            %              keyboard;
            %           end
        end
        
        %%
        function obj = setParametersFromOptimization(obj, params)
            obj.eta = params(end);
            obj.gammA = params(1);
        end
        
        %%
        function optimizationOptions = getOptimizationOptions(obj)
            
            display = 'off';
            
            
            ALG = 'interior-point';
            
            %'trust-region-reflective'
            %'interior-point'
            optimizationOptions = optimset('GradObj',obj.useGradient, 'Display',display,'MaxFunEvals', obj.numOptimizationsDualFunction * 10,...
                'Algorithm', 'interior-point','TolX', 1e-10, 'TolFun', 1e-10, ...
                'MaxIter', obj.numOptimizationsDualFunction,'DerivativeCheck', 'off');
            
            %optimizationOptions = optimset('GradObj',obj.useGradient, 'Display',display,'MaxFunEvals', obj.numOptimizationsDualFunction * 5,...
            %   'Algorithm', ALG,'TolX', 1e-15, 'TolFun', 1e-15, ...
            %  'MaxIter', obj.numOptimizationsDualFunction , 'DerivativeCheck', 'off');
        end
        
        function [g, gD] = dualFunctionPosEta(obj, params)
            params(end) = exp(params(end));
            [g, gD] = obj.dualFunction(params);
        end
        %%
        function [g, gD] = dualFunction(obj, params)
            %  obj.deravitivecheck(params);
            
            eta = params(end); % Lagrangian multiplier
            beta = params(1); % Lagrangian multiplier
            
            distEntropy= obj.entropyBeta; %Current entropy
            
            dimQ = size(obj.Q,1);
            
            F = eta*obj.Qinv - 2 * obj.Raa;
            
            tempSigma = F / (eta + beta);
            [~,p] = chol(tempSigma);
            
            gD = zeros(size(params));
            
            if(any(eig(tempSigma)<=0))
                
                g = Inf;
                gD = Inf * ones(size(gD));
                return;
                
            end
            
            fc = eta*(obj.Q\obj.bias)+obj.ra;
            Fs = eta*(obj.Q\obj.K) + 2 * obj.Ras;
            
            %try
            detTerm = 2*sum(log(diag(chol(F))));
            % catch
            %    detTerm = sum(log(eig(F)));
            %end
            logDetF = dimQ*log((eta+beta) * 2 * pi)-detTerm; % log |2 pi F^-1 (eta + beta)|
            %logDetF = dimQ*log((eta+beta) * 2 * pi)-2*sum(log(diag(chol(F)))); % log |2 pi F^-1 (eta + beta)|
            logDetQ = dimQ*log(2 * pi) + obj.logDetQ; % log |2 pi Q|
            
            
            %desiredEntropy = (obj.entropyBetaDiscount*(distEntropy - obj.minEntropy) + obj.minEntropy) ;
            % desiredEntropy = distEntropy - 0.2;
            
            %             if((desiredEntropy - obj.minEntropy) < 10)
            %
            %                 obj.minEntropy = obj.minEntropy -1 ;
            %
            %             end
            %desiredEntropy = distEntropy-0.2;
            
            constantPart = eta*obj.epsilonAction -(beta * obj.desiredEntropy)+obj.r0 - 0.5 * eta * obj.bias'*(obj.Q\obj.bias) ...
                + 0.5 * fc'*(F\fc) - 0.5 * eta * logDetQ + 0.5 * (eta + beta) * logDetF;
            
            linearPart = 0;
            squaredPart = 0;
            
            if (~isempty(obj.Rss))
                muS = obj.muState;
                linearPart = muS' * (obj.rs - eta * obj.K' * (obj.Q \ obj.bias) + Fs' * (F \ fc));
                M = obj.Rss - 0.5 * eta * obj.K' * (obj.Q \ obj.K) + 0.5 * Fs' * (F \ Fs);
                squaredPart = muS' * M * muS + trace(M * obj.SigmaState);
                
            end
            g = constantPart + linearPart + squaredPart;
            
            
            if (nargout > 1)
                
                %dCovOmega = tempSigma/(eta + beta);
                %dCovEta = dCovOmega - tempSigma*obj.Qinv*dCovOmega;
                
                %dMuEta = dCovOmega * obj.Q\obj.bias - obj.Qinv * dCovOmega * mu;
                
                
                
                %detCov = 2*sum(log(diag(chol(tempSigma))));
                
                %dOmega = 0.5 * (detCov +  dimQ - 2*beta + dimQ*log(2*pi));
                
                %dEta = 0.5 * (detCov +  dimQ - trace(tempSigma*obj.Qinv) + obj.ra' * dMuEta + obj.bias' * obj.Qinv * mu+ ...
                %   eta*obj.bias' * obj.Qinv*dMuEta + 2*obj.epsilonAction - obj.logDetQ - obj.bias'*(obj.Q\obj.bias));
                
                
                invF = inv(F);
                
                invFD = -1 * (invF'*(obj.Q\invF));
                fcD = obj.Q\obj.bias ;
                
                %dEta = obj.epsilonAction - 0.5 * obj.bias'*(obj.Q\obj.bias) + 0.5 *(fc'*invFD*fc) + fc'*(F \ fcD) - 0.5 * logDetQ ...
                %   + 0.5 * logDetF + 0.5 * (eta+beta) * trace( (1/(eta+beta)) + F' * invFD ) ;
                
                dEta = obj.epsilonAction - 0.5 * obj.bias'*fcD  + 0.5 *(fc'*invFD*fc) + fc'*(F \ fcD) - 0.5 * logDetQ ...
                    + 0.5 * logDetF - 0.5 * (eta+beta) * trace(F\obj.Qinv) + 0.5*dimQ;
                
                gD(end) = dEta;
                gD(1) = -obj.desiredEntropy + 0.5*dimQ + 0.5 * logDetF ;
                
            end
            
            
            
            if (isnan(g) || any(isnan(gD)) || ~isreal(g))
                g = inf;
                gD = zeros(size(params));
            end
            
        end
        %%
        function [g, gD] = dualFunction2(obj, params)
            eta = params(end); % Lagrangian multiplier
            beta = params(1); % Lagrangian multiplier
            
            distEntropy= obj.entropyBeta; %Current entropy
            
            dimQ = size(obj.Q,1);
            
            F = eta*obj.Qinv - 2 * obj.Raa;
            
            tempSigma = F / (eta + beta);
            [~,p] = chol(tempSigma);
            
            gD = zeros(size(params));
            
            if(any(eig(tempSigma)<=0))
                
                g = Inf;
                gD = Inf * ones(size(gD));
                return;
                
            end
            
            fc = eta*(obj.Q\obj.bias)+obj.ra;
            Fs = eta*(obj.Q\obj.K) + 2 * obj.Ras;
            
            %try
            detTerm = 2*sum(log(diag(chol(F))));
            % catch
            %    detTerm = sum(log(eig(F)));
            %end
            logDetF = dimQ*log((eta+beta) * 2 * pi)-detTerm; % log |2 pi F^-1 (eta + beta)|
            %logDetF = dimQ*log((eta+beta) * 2 * pi)-2*sum(log(diag(chol(F)))); % log |2 pi F^-1 (eta + beta)|
            logDetQ = dimQ*log(2 * pi) + obj.logDetQ; % log |2 pi Q|
            
            
            %desiredEntropy = (obj.entropyBetaDiscount*(distEntropy - obj.minEntropy) + obj.minEntropy) ;
            desiredEntropy = distEntropy - 0.2;
            
            %             if((desiredEntropy - obj.minEntropy) < 10)
            %
            %                 obj.minEntropy = obj.minEntropy -1 ;
            %
            %             end
            %desiredEntropy = distEntropy-0.2;
            
            constantPart = eta*obj.epsilonAction -(beta * obj.desiredEntropy)+obj.r0 - 0.5 * eta * obj.bias'*(obj.Q\obj.bias) ...
                + 0.5 * fc'*(F\fc) - 0.5 * eta * logDetQ + 0.5 * (eta + beta) * logDetF;
            
            linearPart = 0;
            squaredPart = 0;
            
            if (~isempty(obj.Rss))
                muS = obj.muState;
                linearPart = muS' * (obj.rs - eta * obj.K' * (obj.Q \ obj.bias) + Fs' * (F \ fc));
                M = obj.Rss - 0.5 * eta * obj.K' * (obj.Q \ obj.K) + 0.5 * Fs' * (F \ Fs);
                squaredPart = muS' * M * muS + trace(M * obj.SigmaState);
                
            end
            g = constantPart + linearPart + squaredPart;
            
            
            if (nargout > 1)
                
                %dCovOmega = tempSigma/(eta + beta);
                %dCovEta = dCovOmega - tempSigma*obj.Qinv*dCovOmega;
                
                %dMuEta = dCovOmega * obj.Q\obj.bias - obj.Qinv * dCovOmega * mu;
                
                
                
                %detCov = 2*sum(log(diag(chol(tempSigma))));
                
                %dOmega = 0.5 * (detCov +  dimQ - 2*beta + dimQ*log(2*pi));
                
                %dEta = 0.5 * (detCov +  dimQ - trace(tempSigma*obj.Qinv) + obj.ra' * dMuEta + obj.bias' * obj.Qinv * mu+ ...
                %   eta*obj.bias' * obj.Qinv*dMuEta + 2*obj.epsilonAction - obj.logDetQ - obj.bias'*(obj.Q\obj.bias));
                
                
                invF = inv(F);
                
                invFD = -1 * (invF'*(obj.Q\invF));
                fcD = obj.Q\obj.bias ;
                
                %dEta = obj.epsilonAction - 0.5 * obj.bias'*(obj.Q\obj.bias) + 0.5 *(fc'*invFD*fc) + fc'*(F \ fcD) - 0.5 * logDetQ ...
                %   + 0.5 * logDetF + 0.5 * (eta+beta) * trace( (1/(eta+beta)) + F' * invFD ) ;
                
                dEta = obj.epsilonAction - 0.5 * obj.bias'*fcD  + 0.5 *(fc'*invFD*fc) + fc'*(F \ fcD) - 0.5 * logDetQ ...
                    + 0.5 * logDetF - 0.5 * (eta+beta) * trace(F\obj.Qinv) + 0.5*dimQ;
                
                gD(end) = dEta;
                gD(1) = -obj.desiredEntropy + 0.5*dimQ + 0.5 * logDetF ;
                
                gD = gD';
            end
            
            
            
            if (isnan(g) || any(isnan(gD)) || ~isreal(g))
                g = inf;
                gD = zeros(size(params));
            end
            
        end
        
        %%
        function [] = deravitivecheck(obj, params)
            
            epsilon = 0.0001;
            
            dg = ones(length(params),1);
            
            for(i=1:length(params))
                
                params1 = params;
                params2 = params;
                
                params1(i) = params(i) + epsilon;
                params2(i) = params(i) - epsilon;
                
                [g1, ~]  = obj.dualFunction2(params1);
                [g2, ~]  = obj.dualFunction2(params2);
                
                dg(i) = (g1-g2) / (2*epsilon);
                
                
            end
            
            [~, dgAnalytic]  = obj.dualFunction2(params);
            
            dg - dgAnalytic ;
            
        end
        %%
        
        function [updatePolicy] = updatePolicy(obj)
            
            eta = obj.eta;
            
            %quadratic part
            F = eta*obj.Qinv - 2 * obj.Raa;
            %linear part (in s)
            Fs = eta*(obj.Q\obj.K) + 2 * obj.Ras;
            %constant part
            fc = eta*(obj.Q\obj.bias)+obj.ra;
            
            %scale with variance
            Fs = F \ Fs;
            fc = F \ fc;
            
            tmpSigma = inv(F / (obj.eta + obj.gammA));
            
            obj.divKL = 0.5 * (sum(log(eig(obj.Q))) - sum(log(eig(tmpSigma))) + trace(obj.Q \ tmpSigma)); ... %  const terms of KL
                obj.divCov = obj.divKL- 0.5 * size(obj.bias, 1);
            
            constantDiff = obj.bias - fc;
            linearDiff = obj.K - Fs;
            muS = obj.muState;
            sigmaS = obj.SigmaState;
            obj.divKL = obj.divKL + 0.5 * (constantDiff' * obj.Qinv * constantDiff) - 0.5 * size(obj.bias, 1);
            obj.divMean = 0.5 * (constantDiff' * obj.Qinv * constantDiff);
            
            obj.rewardAfterUpdate = fc' * obj.Raa * fc + trace(obj.Raa *  tmpSigma) + fc' * obj.ra + obj.r0;
            
            bias = fc;
            
            cholCov = chol(tmpSigma);
            v = 2*sum(log(diag(cholCov))); %lodDet Policy covariance
            
            obj.entropyAfter =0.5 * length(bias)*(1+log(2*pi)) + 0.5*v; %1/2*(log(((2*pi*exp(1))^length(bias)) * det(cov))/log(exp(1)));
            
            
            
            %  if (obj.divKL < obj.epsilonAction * 1.2 && abs(obj.entropyAfter - obj.desiredEntropy) < 0.2)
            obj.policy.setCovariance(tmpSigma);
            obj.policy.setWeightsAndBias(Fs, fc);
            updatePolicy = true;
            % else
            %     warning('KL %f or entropy %f too high, no update applied!\n', obj.divKL, obj.entropyAfter);
            %     updatePolicy = false;
            % end
            
            
            
            if (~isempty(muS))
                obj.divKL = obj.divKL + 0.5 * ( 2 * muS' * linearDiff' * obj.Qinv * constantDiff + ...
                    muS' * linearDiff' * obj.Qinv * linearDiff * muS + trace(linearDiff' * obj.Qinv * linearDiff * sigmaS));
            end
            
            
        end
    end
end