classdef EpisodicPower < Learner.WeightedML.RLByWeightedML
    properties
        divKL
    end
    
    properties (SetObservable,AbortSet)
        temperatureScalingPower = 10.0;
    end
    
    methods (Static)
        function [learner] = CreateFromTrial(trial)
            learner = Learner.EpisodicRL.EpisodicPower(trial.dataManager, trial.parameterPolicyLearner);
        end
        
        function [learner] = CreateFromTrialKnowsNoise(trial)
            trial.transitionFunction.registerControlNoiseInData();
            learner = Learner.EpisodicRL.EpisodicPower(trial.dataManager, trial.policyLearner);
            learner.addDataPreprocessor(DataPreprocessors.NoiseActionPreprocessor(trial.dataManager));           
            trial.policyLearner.setOutputVariableForLearner('actionsWithNoise');

        end
    end
        
    % Class methods
    methods
        function obj = EpisodicPower(dataManager, policyLearner, varargin)            
            obj = obj@Learner.WeightedML.RLByWeightedML(dataManager, policyLearner, varargin{:});            
                       
            obj.linkProperty('temperatureScalingPower');            
        end                        
        
        function [weighting] = computeWeighting(obj, rewards)            
            maxQ = max(rewards);
            minQ = min(rewards);
            
            weighting = exp(obj.temperatureScalingPower * (rewards - maxQ) / ((maxQ - minQ) + 10^-6));
            weighting = weighting / sum(weighting);
            
            obj.divKL = obj.getKLDivergence(ones(size(weighting)), weighting);
        end
        
        function [] = printMessage(obj, data)
            fprintf('divKL %f\n', obj.divKL);            
        end
                
    end    
end

