classdef NESLearner2 < Learner.RLLearner & Data.DataManipulator
    
    
    properties
        
        policy;
        
        
        
        shape;
        weights;
        fit;
        rewardName;
        parameterName;
        
        
    end
    
    properties(SetObservable, AbortSet)
        learnRateNESMeans;
        learnRateNESSigmas;
        L;
    end
    
    methods (Static)
        function [learner] = CreateFromTrial(trial)
            learner = Learner.EpisodicRL.NESLearner2(trial.dataManager, trial.parameterPolicy, 'returns', 'parameters');
        end
    end
    
    
    % Class methods
    methods
        function obj = NESLearner2(dataManager, policy, rewardName, parameterName)
            obj = obj@Learner.RLLearner();
            obj = obj@Data.DataManipulator(dataManager);
            obj.policy = policy;
            
            if (~exist('rewardName', 'var'))
                rewardName = 'returns';
            end
            
            if (~exist('parameterName', 'var'))
                parameterName = 'parameters';
            end
            
            
            
            obj.rewardName = rewardName;
            obj.parameterName = parameterName;
            
            
            minRange = obj.dataManager.getMinRange(obj.parameterName);
            maxRange = obj.dataManager.getMinRange(obj.parameterName);
            
            
            dimParameters = dataManager.getNumDimensions(parameterName);
            
            obj.L = 4+3*floor(log(dimParameters));
           
            obj.learnRateNESMeans=1;
            obj.learnRateNESSigmas=0.5*min(1.0/dimParameters,0.25);
            %obj.learnRateNESSigmas = 0.1 ;
         
            
            obj.linkProperty('learnRateNESMeans');
            obj.linkProperty('learnRateNESSigmas');
            obj.linkProperty('L');
            %obj.fpt
            %obj.xopt
            Common.Settings().setProperty('newSamplesEpisode', obj.L);
           % Common.Settings().setProperty('initialSamplesEpisode', 0);
            Common.Settings().setProperty('maxSamples', obj.L);
            
            obj.registerLearningFunction();
            
        end
        
        function obj = initObject(obj)
            % abstract init function
            
            obj.initObject@Learner.RLLearner();
            
            
            
            %obj.settings.setParameter('newSamples', lambda);
            %obj.settings.setParameter('initialSamples', 0);
            %obj.settings.setParameter('maxSamples', lambda);
        end
        
        function [] = registerLearningFunction(obj)
            obj.addDataManipulationFunction('computePolicyUpdate', {obj.rewardName, obj.parameterName}, {});
        end
        
        
        
        function [] = computePolicyUpdate(obj, rewards, parameters)
            
            obj.shape = max(0.0, log(obj.L/2+1.0)-log(1:obj.L));
            obj.shape = obj.shape / sum(obj.shape);
            obj.weights = zeros(1,obj.L);
            obj.fit = zeros(1,obj.L);
            d = obj.dataManager.getNumDimensions(obj.parameterName);
            expA = obj.policy.cholA;
            A = logm(obj.policy.cholA);
            x = (obj.policy.bias);
            X = parameters';
            Z = expA \ (X-repmat(x,1,obj.L));
            %Z = randn(d,L); X = repmat(x,1,L)+expA*Z;
            obj.fit = -rewards';
            [~, idx] = sort(obj.fit);
            obj.weights(idx) = obj.shape;
       
            % step 3: compute the gradient for C and x
            G = (repmat(obj.weights,d,1).*Z)*Z' - sum(obj.weights)*eye(d);
            dx = obj.learnRateNESMeans * expA * (Z*obj.weights');
            dA = obj.learnRateNESSigmas * G;
            % The CMA-ES optimizer is a minimizer, we have to take the
            % negative reward function
            x = x + dx;
            A = A + dA;
            
            obj.policy.setBias(x);
            obj.policy.setSigma(expm(A));
            
        end
        
        function [] = updateModel(obj, data)
            obj.callDataFunction('computePolicyUpdate', data);
        end
        
        function [] = printMessage(obj, data, results, model)
        end
        
        
        % methods
    end
end